
from modules.fqlAstBuilder.ast import FqlAstBuilder
from modules.fqlAstTranspiler.postgreTranspiler import FqlAstPostgreSqlTranspiler


class TestFqlAstPostgreSqlTranspiler:
    def test_fqlStatementTranspiler_createForm(self):
        code = "create form Person(name TEXT, age NUMBER, project references Projects)"
        fqlAstBuilder = FqlAstBuilder.fqlAstFromCode(code)
        model = fqlAstBuilder.model
        postgreTanspiler = FqlAstPostgreSqlTranspiler(model)
        result = postgreTanspiler.generate()
        assert result == "CREATE TABLE Person(FQL_ID SERIAL PRIMARY KEY,name VARCHAR(511),age INTEGER,project INTEGER,FOREIGN KEY(project) REFERENCES Projects(FQL_ID));"

    def test_fqlStatementTranspiler_getCase(self):
        code = "get Person"
        fqlAstBuilder = FqlAstBuilder.fqlAstFromCode(code)
        model = fqlAstBuilder.model
        postgreTanspiler = FqlAstPostgreSqlTranspiler(model)
        result = postgreTanspiler.generate()
        assert result == "SELECT * FROM Person;"

    def test_fqlStatementTranspiler_getCaseWithConditions(self):
        code = "get Person with name='Leandro' and age=20"
        fqlAstBuilder = FqlAstBuilder.fqlAstFromCode(code)
        model = fqlAstBuilder.model
        postgreTanspiler = FqlAstPostgreSqlTranspiler(model)
        result = postgreTanspiler.generate()
        assert result == "SELECT * FROM Person WHERE name='Leandro' and age=20;"
        
    def test_fqlStatementTranspiler_createNew(self):
        code = "create Person (name='Leandro', age=23)"
        fqlAstBuilder = FqlAstBuilder.fqlAstFromCode(code)
        model = fqlAstBuilder.model
        postgreTanspiler = FqlAstPostgreSqlTranspiler(model)
        result = postgreTanspiler.generate()
        assert result == "INSERT INTO Person (name,age) VALUES ('Leandro',23);"



class AstTranspiler:
    def __init__(self, astModel) -> None:
        self.model = astModel

    def generate(self):
        NotImplemented()


class FqlAstTranspiler(AstTranspiler):
    def __init__(self, fqlAstModel) -> None:
        super().__init__(fqlAstModel)


class FqlAstSqlTranspiler(FqlAstTranspiler):
    def __init__(self, fqlAstModel) -> None:
        super().__init__(fqlAstModel)



from core import getPostgreCodeFromFqlCode

postgreCode1 = getPostgreCodeFromFqlCode("create form Person(name TEXT, age NUMBER)")
postgreCode2 = getPostgreCodeFromFqlCode("create form Person(name TEXT, age NUMBER, Product references Products)")

postgreCode3 = getPostgreCodeFromFqlCode("create new Person (name='Leandro', age=23, isCold=True)")

postgreCode4 = getPostgreCodeFromFqlCode("get Person")
postgreCode5 = getPostgreCodeFromFqlCode("get Person with name='Leandro'")
postgreCode6 = getPostgreCodeFromFqlCode("get Person with name='Leandro' and age=23")




print(postgreCode1)
print(postgreCode2)
print()
print(postgreCode3)
print()
print(postgreCode4)
print(postgreCode5)
print(postgreCode6)




from fqllang.dsl.utils import getClassName
from fqllang.modules.fqlAstTranspiler.transpiler import FqlAstSqlTranspiler
from fqllang.modules.postgreSqlCodeGenerator.create import CreatePostgreSql
from fqllang.modules.postgreSqlCodeGenerator.insert import PreparedInsertPostgreSql
from fqllang.modules.postgreSqlCodeGenerator.select import SelectPostgreSql, SelectWithCriteriaPostgreSql
from fqllang.modules.sqlCodeGenerator.elements import Column, ColumnList, Condition, Criteria, DataConstraint, DataType, Field, FieldList, Reference, ValueList, convertFqlDataTypeToPostgreSqlDataType


class FqlAstPostgreSqlTranspiler(FqlAstSqlTranspiler):
    def __init__(self, fqlAstModel) -> None:
        super().__init__(fqlAstModel)

    def generate(self):
        return self.fqlStatementTranspiler(self.model)

    def __str__(self) -> str:
        return self.generate()

    def fqlStatementTranspiler(self, model):
        className = getClassName(model)
        if className == "CreateForm":
            return self.createFormTranspiler(model).generate()
        elif className == "CreateNew":
            return self.createNewTranspiler(model).generate()
        elif className == "ShowForms":
            return self.showFormsTranspiler(model).generate()
        elif className == "GetCase":
            return self.getCaseTranspiler(model).generate()
        raise Exception("Modelo incorrecto")

    def createFormTranspiler(self, createForm):
        formName = createForm.formName

        dataSpecList = self.dataSpecListTranspiler(createForm.dataSpecList)

        keyField = Field('FQL_ID', DataType.serial, DataConstraint.primaryKey)
        
        fieldList = FieldList(keyField, *dataSpecList)

        return CreatePostgreSql(formName, ColumnList.emptyColumnList(), fieldList)

    def dataSpecListTranspiler(self, dataSpecList):
        dataSpecListTranspiled = map(lambda dataSpec: self.dataSpecTranspiler(
            dataSpec), dataSpecList)
        return list(dataSpecListTranspiled)
    
    def dataSpecTranspiler(self, dataSpec):
        className = getClassName(dataSpec)
        if className == "DataDefinition":
            return self.dataDefinitionTranspiler(dataSpec)
        elif className == "DataReference":
            return self.dataReferenceTranspiler(dataSpec)

    def dataDefinitionTranspiler(self, dataDefinition):
        dataName = dataDefinition.dataName
        dataType = convertFqlDataTypeToPostgreSqlDataType(
            dataDefinition.dataType)
        notNull = dataDefinition.notNull
        unique = dataDefinition.unique
        field = Field(dataName, dataType)
        if notNull:
            field.addDataConstraint(DataType.notNull)
        if unique:
            field.addDataConstraint(DataType.unique)
        return field

    def dataReferenceTranspiler(self, dataReference):
        dataName = dataReference.dataName
        refCardinality = dataReference.refCardinality
        refPath = dataReference.refPath
        basicField = Field(dataName, DataType.integer)
        reference = Reference(basicField, refPath)
        return reference

    def createNewTranspiler(self, createNew):
        formName = createNew.formName
        columns, values = self.createNewValuesTranspiler(createNew.createNewValues)
        return PreparedInsertPostgreSql(formName, columns, values)

    def createNewValuesTranspiler(self, createNewValues):
        keyValueEntryList = createNewValues.keyValueEntryList
        keyValueList = [self.keyValueEntryTranspiler(keyValueEntry) for keyValueEntry in keyValueEntryList]
        columns = []
        values = []
        for key, value in keyValueList:
            columns.append(key)
            values.append(value)
        return ColumnList.createColumnListByElements(*columns), ValueList.createColumnListByElements(*values)
        
    def keyValueEntryTranspiler(self, keyValueEntry):
        className = getClassName(keyValueEntry)
        if className == "KeyValueSimple":
            return self.keyValueSimpleTranspiler(keyValueEntry)
        elif className == "KeysSeparatesWithDots":
            pass
        
    def keyValueSimpleTranspiler(self, keyValueSimple):
        key = keyValueSimple.key
        valueTerminal = keyValueSimple.valueTerminal
        return key, valueTerminal

    def getCaseTranspiler(self, getCase):
        formName = getCase.formName
        conditions = self.extractConditions(getCase)
        if conditions:
            criteria = self.getCaseConditionsTranspiler(conditions)
            return SelectWithCriteriaPostgreSql(formName, ColumnList.emptyColumnList(), criteria)
        else:
            return SelectPostgreSql(formName)
        # extendedLabelsList = self.extractExtendedLabelsList(getCase)

    def getCaseConditionsTranspiler(self, conditions):
        conditionList = []
        for condition in conditions:
            className = getClassName(condition)
            if className == "ConditionTerminal":
                attribute = condition.attribute
                valueTerminal = condition.valueTerminal
                conditionList.append(Condition(attribute, valueTerminal))
        return Criteria(*conditionList)

    def extractConditions(self, getCase):
        try:
            return getCase.conditions
        except:
            return []

    # def extractExtendedLabelsList(self, getCase):
    #     try:
    #         return getCase.extendedLabelsList
    #     except:
    #         return []

    def showFormsTranspiler(self, showForms):
        pass



from fqllang.modules.sqlCodeGenerator.base import CreateSql
from fqllang.modules.sqlCodeGenerator.elements import ColumnList, FieldList


class CreatePostgreSql(CreateSql):
    def __init__(self, tableName: str, columns: ColumnList, fields: FieldList) -> None:
        super().__init__(tableName, columns)
        self.fields = fields

    def generate(self):
        result = f"CREATE TABLE {self.tableName}("
        result += f"{self.fields}"
        result += f");"
        return result



from fqllang.modules.sqlCodeGenerator.base import InsertSql, PreparedInsertSql
from fqllang.modules.sqlCodeGenerator.elements import Field, ColumnList, ValueList


class InsertPostgreSql(InsertSql):
    def __init__(self, tableName: str, columns: ColumnList, fields: list[Field]) -> None:
        super().__init__(tableName, columns, fields)

    def _valuesList(self, fields: list[Field], columns: ColumnList):
        pass


class PreparedInsertPostgreSql(PreparedInsertSql):
    def __init__(self, tableName: str, columns: ColumnList, values: ValueList) -> None:
        super().__init__(tableName, columns)
        self.values = values

    def _placeholderList(self, columns: ColumnList):
        pass
    
    def generate(self):
        return f"INSERT INTO {self.tableName} {self.columns} VALUES {self.values};"
    
    def __str__(self) -> str:
        return self.generate()





from dsl.fql.base import Fql
from dsl.utils import getClassName

class Form:
    def __init__(self, name, *fields) -> None:
        self.name = name
        self.fields = fields
        
    def generate(self):
        return CreateForm(self.name, *self.fields).generate()
    
    def __str__(self):
        return self.generate()
    
class FormList:
    def __init__(self, *forms) -> None:
        self.forms = forms
        
    def __iter__(self):
        for form in self.forms:
            yield form
            
    def generate(self):
        strGenerated = map(lambda form: form.generate(), self)
        return "\n".join(strGenerated)
    
    def __str__(self) -> str:
        return self.generate()


class CreateForm(Fql):
    def __init__(self, formName, *dataSpecs) -> None:
        self.formName = formName
        self.dataSpecs = dataSpecs


class DataSpec(Fql):
    def __init__(self, dataName) -> None:
        self.dataName = dataName


class DataDefinition(DataSpec):
    def __init__(self, dataName, dataType, primaryKey=None, notNull=None, unique=None) -> None:
        super().__init__(dataName)
        self.dataType = dataType
        self.notNull = notNull
        self.unique = unique
        self.primaryKey = primaryKey


class DataReference(DataSpec):
    def __init__(self, dataName, refPath, refCardinality=None) -> None:
        super().__init__(dataName)
        self.refPath = refPath
        self.refCardinality = refCardinality


class RefCardinality(Fql):
    def __init__(self, cardinality1, cardinality2=None) -> None:
        super().__init__()
        self.cardinality1 = cardinality1
        self.cardinality2 = cardinality2



from fqllang.modules.sqlCodeGenerator.elements import Field, Criteria, ColumnList


class Sql:
    def __init__(self, tableName: str, columns: ColumnList) -> None:
        self.tableName = tableName
        if columns:
            self.columns = columns
        else:
            self.columns = ColumnList.emptyColumnList()

    def generate(self):
        NotImplemented()


class CreateSql(Sql):
    def __init__(self, tableName: str, columns: ColumnList) -> None:
        super().__init__(tableName, columns)


class SelectSql(Sql):
    def __init__(self, tableName: str, columns: ColumnList) -> None:
        super().__init__(tableName, columns)


class InsertSql(Sql):
    def __init__(self, tableName: str, columns: ColumnList, fields: list[Field]) -> None:
        super().__init__(tableName, columns)

    def _valuesList(self, fields: list[Field], columns: ColumnList):
        pass


class SelectWithCriteriaSql(Sql):
    def __init__(self, tableName: str, columns: ColumnList, criteria: Criteria) -> None:
        super().__init__(tableName, columns)
        self.criteria = criteria


class SelectWithMatchSql(Sql):
    def __init__(self, tableName: str, columns: ColumnList, pattern: str) -> None:
        super().__init__(tableName, columns)


class FindByKeySql(Sql):
    def __init__(self, tableName: str, columns: ColumnList, keyColumn: str, keyValue: str):
        super().__init__(tableName, columns)


class PreparedInsertSql(Sql):
    def __init__(self, tableName: str, columns: ColumnList) -> None:
        super().__init__(tableName, columns)

    def _placeholderList(self, columns: ColumnList):
        pass



from fqllang.dsl.utils import getGrammar
from fqllang.modules.fqlAstBuilder.ast import FqlAstBuilder
from fqllang.modules.fqlAstTranspiler.postgreTranspiler import FqlAstPostgreSqlTranspiler

GRAMMAR = getGrammar()


def getPostgreCodeFromFqlCode(fqlCode):
    try:
        fqlAstBuilder = FqlAstBuilder.fqlAstFromCode(fqlCode)
        postgreCode = FqlAstPostgreSqlTranspiler(fqlAstBuilder.model)
        return postgreCode
    except:
        raise Exception("Bad Fql Code")


if __name__ == "__main__":
    postgreCode = getPostgreCodeFromFqlCode("get Person with age=5")
    print(postgreCode)




from modules.fqlAstBuilder.ast import FqlAstBuilder
from modules.fqlBuilder.modelAnalizer import ModelAnalizer

def model_from_code(code):
    model = FqlAstBuilder.fqlAstFromCode(code)
    ast = ModelAnalizer(model).analizeModel()
    return ast

code = "create form Person(Project references Projects)"
model_from_code(code)
code = "create form Person(Project references 1 Projects)"
model_from_code(code)
code = "create form Person(Project references 0..1 Projects)"
model_from_code(code)
code = "create form Person(Project references 0..many Projects)"
model_from_code(code)
code = "create form Person(Project references 0..many Projects.*)"
model_from_code(code)
code = "create form Person(Project references 0..many Projects.Animal.*)"
model_from_code(code)
code = "create form Person(Project references 0..many Projects.Animal.(name, age) UNIQUE)"
model_from_code(code)
code = "create form Person(Project references 0..many Projects.Animal.(name, age) TOTALLY UNIQUE)"
model_from_code(code)






from dsl.fql.createForm import Form
from dsl.utils import getClassName
from dsl.fql.base import Field, Reference, RefCardinality, RefConstraint, SimpleId, DataPath, All, DataLabelList, DataLabel


class ModelAnalizer:
    def __init__(self, model) -> None:
        self.model = model
        
    def analizeModel(self):
        className = getClassName(self.model)
        if className == "CreateForm":
            return self.analizeCreateForm(self.model)
            
    def analizeCreateForm(self, createForm):
        formName = createForm.formName
        fields = self.analizeDataSpecList(createForm.dataSpecList)
        return Form(formName, *fields)
    
    def analizeDataSpecList(self, dataSpecList):
        dataSpecIterator = map(self.analizeDataSpec, dataSpecList)
        fields = list(dataSpecIterator)
        return fields
    
    def analizeDataSpec(self, dataSpec):
        className = getClassName(dataSpec)
        if className == "DataDefinition":
            return self.analizeDataDefinition(dataSpec)
        elif className == "DataReference":
            return self.analizeDataReference(dataSpec)
        
    def analizeDataDefinition(self, dataDefinition):
        dataName = dataDefinition.dataName
        dataType = dataDefinition.dataType
        notNull = dataDefinition.notNull
        unique = dataDefinition.unique
        return Field(dataName, dataType, notNull, unique)
    
    def analizeDataReference(self, dataReference):
        dataName = dataReference.dataName
        refCardinality = self.analizeRefCardinality(dataReference.refCardinality)
        refPath = self.analizeRefPath(dataReference.refPath)
        refConstraint = self.analizeRefConstraint(dataReference.refConstraint)
        return Reference(dataName, refCardinality, refPath, refConstraint)
    
    def analizeRefCardinality(self, refCardinality):
        cardinality1 = refCardinality.cardinality1
        cardinality2 = refCardinality.cardinality2
        return RefCardinality(cardinality1, cardinality2)
    
    def analizeRefPath(self, refPath):
        className = getClassName(refPath)
        if className == "SimpleId":
            return self.analizeSimpleId(refPath)
        elif className == "IdDotDataPath":
            return self.analizeIdDotDataPath(refPath)
        
    def analizeSimpleId(self, simpleId):
        return SimpleId(simpleId)
    
    def analizeIdDotDataPath(self, idDotDataPath):
        dataPath = []
        dataPath.append(SimpleId(idDotDataPath.id))
        dataPath.extend(self.analizeDataPath(idDotDataPath.dataPath))
        return DataPath(*dataPath)
    
    def analizeDataPath(self, dataPath):
        className = getClassName(dataPath)
        if className == "SimpleId":
            return self.analizeSimpleId(dataPath)
        elif className == "IdDotDataPath":
            return self.analizeIdDotDataPath(dataPath)
        elif className == "DataLabelList":
            return self.analizeDataLabelList(dataPath.dataLabelList)
        elif className == "All":
            return self.analizeAll(dataPath)
        
    def analizeDataLabelList(self, dataLabelList):
        dataLabels = list(map(self.analizeDataLabel, dataLabelList))
        return DataLabelList(*dataLabels)
    
    def analizeDataLabel(self, dataLabel):
        return DataLabel(dataLabel)
    
    def analizeAll(self, all):
        return All(all)
        
    def analizeRefConstraint(self, refConstraint):
        return RefConstraint(refConstraint)
    



class Fql:
    pass


class Field:
    def __init__(self, name, dataType, *dataConstraints) -> None:
        self.name = name
        self.dataType = dataType
        self.dataConstraints = dataConstraints


class Reference:
    def __init__(self, dataName, refCardinality, refPath, refConstraint) -> None:
        self.dataName = dataName
        self.refCardinality = refCardinality
        self.refPath = refPath
        self.refConstraint = refConstraint


class RefCardinality:
    def __init__(self, cardinality1, cardinality2) -> None:
        self.cardinality1 = cardinality1
        self.cardinality2 = cardinality2


class RefConstraint:
    def __init__(self, constraint) -> None:
        self.constraint = constraint


class SimpleId:
    def __init__(self, id) -> None:
        self.id = id


class DataPath:
    def __init__(self, *dataPath) -> None:
        self.path = dataPath

    def __iter__(self):
        for data in self.dataPath:
            yield data


class DataLabelList:
    def __init__(self, *dataLabels) -> None:
        self.dataLabels = dataLabels


class DataLabel(SimpleId):
    def __init__(self, id) -> None:
        super().__init__(id)


class All:
    def __init__(self, all) -> None:
        self.star = all
